/*
 * ??filename??
 *
 * Copyright 1997 Corel Computer Corp
 *
 */

/* 
 * Support for interrupt services
 *
 */

/* FootBridge translation tables 
	Converts from our IRQ numbers into FootBridge masks
*/
int fb_irq_mask[16] = {
	0		// 0 not used
	2,		// IRQ_SWI 	1
	0xC,		// IRQ_UART_DEBUG
	0,		// 3 not used
	0x10,		// IRQ_TIMER0
	0x20,		// IRQ_TIMER1
	0x40,		// IRQ_TIMER2
	0x80,		// IRQ_WATCHDOG	
	0x100,		// IRQ_ETHER10
	0x200,		// IRQ_ETHER100
	0x400,		// IRQ_VIDCOMP
	0x800, 		// IRQ_EXTERN_IRQ
	0x30000, 	// IRQ_DMA1
	0,		// 13 not used
	0,		// 14 not used
	0xF000000 	// IRQ_PCI_ERR 
};

/* Conversion from our IRQ numbers into PIC mask is easy, because
	there are no gaps. Given an IRQ (n) that is 16<=n<=31,
	if (n<24) lo_mask = 1 << (n-16)
	else hi_mask = 1 << (n-24)
*/

/* In the following code, the irq number is split into three "zones".
	The first zone deals with interrupts sourced by the FootBridge.
	The second and third zones deal with irq's from the PIC (intel-type)
	controller */
void mask_irq(unsigned int irq) {
	unsigned char t;

	// we should never turn of the chaining irq's
	assert (irq!=10);
	assert (irq!=18);

	if (irq < 16)
	        *IRQ_CLEAR = fb_irq_mask[irq];
	else if (irq < 24) {
		t = inb(PIC_MASK_LO);
		t |= (1 << (irq - 16));
		outb(t, PIC_MASK_LO);
	} else {
		t = inb(PIC_MASK_HI);
		t |= (1 << (irq - 24));
		outb(t, PIC_MASK_HI);
	}
}


void unmask_irq(unsigned int irq) {
	unsigned char t;
	if (irq < 16)
	        *IRQ_SET = fb_irq_mask[irq];
	else if (irq < 24) {
		t = inb(PIC_MASK_LO);
		t &= ~(1 << (irq - 16));
		outb(t, PIC_MASK_LO);
	} else {
		t = inb(PIC_MASK_HI);
		t &= ~(1 << (irq - 24));
		outb(t, PIC_MASK_HI);
	}
}

void irq_init_irq(void) {
	unsigned long flags;

	save_flags (flags);
	cli ();
	/* clear all interrupt enables
		except the two chaining ones */
	*IRQ_CLEAR = 0xffffffff - IRQ_EXTERN_IRQ; 
	outb(0xFF - 2, PIC_MASK_LO);
	outb(0xFF, PIC_MASK_HI); 
	restore_flags (flags);
}

/* from what we know of the hardware, build a 32 bit mask
	showing the current enabled irq's
*/
unsigned long get_enabled_irqs(void) {
	unsigned long fb_raw, u, ret_val;
	unsigned char pic_lo, pic_hi;
	
	fb_raw = *IRQ_SHOW ;
	pic_lo = inb(PIC_MASK_LO);
	pic_hi = inb(PIC_MASK_HI);

	// now, run through the fb ones and build
	// a mask of logical bits
	for (u=0; u<16; u++) {
		if (fb_raw & fb_irq_mask[u])
			ret_val |= (1<<u);
	}

	// the PIC irqs are one-for-one with the logicals
	ret_val |= (pic_lo << 16);
	ret_val |= (pic_hi << 24);
	
	return ret_val;
}
