/*
 * linux/arch/arm/lib/irqs-vnc.S
 *
 * Copyright (C) 1997 Corel Computer Corp
 *
 * Interrupt wrappers - handles disabling & re-enabling interrupts
 *
 * Interrupt numbers (in r0) below 16 refer to footbridge interrupts
 * 	these are easy to handle
 * Interrupt numbers above between 16 and 23 are PIC_LO 
 */

#include <asm/assembler.h>
#include <asm/arch/hardware.h>		// for interrupt hardware

@ IRQ stubs

@ IRQ stubs entered with:
@ r0 = IRQ number
@ r1 = regs
@ return 0 for normal irq, 1 for fast

		.text
/*
 * Disable interrupt & print a message
 */
LC0:		.long	intr_count
LC1:		.long	fb_irq_mask

		.globl	bad_IRQ_interrupt
bad_IRQ_interrupt:
		mov	r9, lr
		mov	r5, #1
		mov	r5, r5, lsl r0
		strb	r5, [r4]
		bl	bad_IRQ
		mov	r0, #0
		mov	pc, r9

		.globl	fast_IRQ_interrupt
		@ r0 irq number
		@ r1 *regs
		@ r2
		@ r3
		@ r4 cpsr
		@ r5 &footbridge or &PIC_HI/_LO
		@ r6 fb irq mask or PIC mask
		@ r7 &intr_count
		@ r8 intr_count
		@ r9 return address
fast_IRQ_interrupt:
		mov	r9, lr
	@ which zone: footbridge or PIC?
		cmp	r0, #16
		bge	1f

	@ first zone: fb irq's
	@ block the irq
		mov	r5, #IO_BASE_ARM_CSR
		ldr	r6, LC1
		ldr 	r6, [r6, r0, lsl #2]
		str	r6, [r5, #IRQ_ENABLE_CLEAR]
	@ incrment the intr_count
		ldr	r7, LC0
		ldr	r8, [r7]
		add	r2, r8, #1
		str	r2, [r7]
	@ call the handler
		bl	do_IRQ
	@ cli
		mrs	r4, cpsr
		orr	r4, r4, #I_BIT
		msr	cpsr, r4
	@ and re-enable it
		str	r6, [r5, #IRQ_ENABLE_SET]
	@ intr_count--
		str	r8, [r7]
		@ return 1
		mov	r0, #1
		mov	pc, r9	

	@ second zone: lo PIC interrupt
1:		
@	mov	r4, r0
@	mov r0, #'y'
@	bl _putchar
@	mov r0, r4
@	add r0, r0, #'0'
@	bl _putchar
@	mov r0, r4
	@ resolve whether we deal with lo PIC or hi PIC
		mov	r2, #IO_BASE
		cmp	r0, #24
		addge	r5, r2, #PIC_HI
		addlt	r5, r2, #PIC_LO
		subge	r3, r0, #24
		sublt	r3, r0, #16
	@ block the interrupt
		mov 	r6, #1
		mov	r6, r6, lsl r3
		mov	r7, #IO_BASE_ARM_CSR
		mov	r4, #0x40000000
//		str	r4, [r7,  #PCI_ADDR_EXT]
		ldrb	r3, [r5, #1]
		orr	r3, r3, r6
		strb	r3, [r5, #1]
		mov	r4, #0
//		str	r4, [r7,  #PCI_ADDR_EXT]
		mov	r3, #0x20
		strb	r3, [r5]
		b	9f
9:		strgeb	r3, [r2, #PIC_LO]
	@ increment intr_count
@ mov r2, r0
@ add r0, r0, #0x40
@ bl _putchar
@ mov r0, r2
		ldr	r7, LC0
		ldr	r8, [r7]
		add	r2, r8, #1
		str	r2, [r7]
	@ call the handler
		bl	do_IRQ
	@ cli
		mrs	r4, cpsr
		orr	r4, r4, #I_BIT
		msr	cpsr, r4
	@ instr_count--
		str	r8, [r7]
	@ and re-enable it
		mov	r7, #IO_BASE_ARM_CSR
		mov	r4, #0x40000000
//		str	r4, [r7,  #PCI_ADDR_EXT]
		ldrb	r3, [r5, #1]
		bic	r3, r3, r6
		strb	r3, [r5, #1]
		mov	r4, #0
//		str	r4, [r7,  #PCI_ADDR_EXT]
	@ and return
		mov	r0, #1
		mov	pc, r9

		.globl	IRQ_interrupt
IRQ_interrupt:
		mov	r9, lr				@ lr in case we get interrupted
		cmp	r0, #16				@ disable this particular interrupt
		bge	1f				@ skip to PIC type interrupts
		mov	r5, #IO_BASE_ARM_CSR		@ handle footbridge int
		ldr	r6, LC1				@ init in irq.c
		ldr	r6, [r6, r0, lsl #2]		@ get the bit mask
		str	r6, [r5, #IRQ_ENABLE_CLEAR]	@ block further ints of this type
		ldr	r7, LC0
		ldr	r8, [r7]
		add	r2, r8, #1			@increment the intr_count
		str	r2, [r7]

		mrs	r4, cpsr			@ Enable interrupts
		bic	r2, r4, #I_BIT
		msr	cpsr, r2
		bl	do_IRQ				@ call the handler
		msr	cpsr, r4
		str	r8, [r7]			@ restore intr_count
		str	r6, [r5, #IRQ_ENABLE_SET]	@ turn it back on
		mov	r0, #0				@ and return
		mov	pc, r9		

1:	@ resolve whether we deal with lo PIC or hi PIC
		mov	r2, #IO_BASE
		cmp	r0, #24
		addge	r5, r2, #PIC_HI
		addlt	r5, r2, #PIC_LO
		subge	r3, r0, #24
		sublt	r3, r0, #16
	@ block the interrupt
		mov 	r6, #1
		mov	r6, r6, lsl r3
		mov	r7, #IO_BASE_ARM_CSR
		mov	r4, #0x40000000
//		str	r4, [r7,  #PCI_ADDR_EXT]
		ldrb	r3, [r5, #1]
		orr	r3, r3, r6
		strb	r3, [r5, #1]
		mov	r4, #0
//		str	r4, [r7,  #PCI_ADDR_EXT]
		mov	r3, #0x20
		strb	r3, [r5]
		strgeb	r3, [r2, #PIC_LO]
	@ increment intr_count
		ldr	r7, LC0
		ldr	r8, [r7]
		add	r2, r8, #1
		str	r2, [r7]
	@ Enable interrupts
		mrs	r4, cpsr		
		bic	r2, r4, #I_BIT
		msr	cpsr, r2
	@ call the handler
		bl	do_IRQ
	@ cli
		msr	cpsr, r4	@?? is this enough? just to restore, instead of a full cli?
	@ instr_count--
		str	r8, [r7]
	@ and re-enable it
		mov	r7, #IO_BASE_ARM_CSR
		mov	r4, #0x40000000
//		str	r4, [r7,  #PCI_ADDR_EXT]
		ldrb	r3, [r5, #1]
		bic	r3, r3, r6
		strb	r3, [r5, #1]
		mov	r4, #0
//		str	r4, [r7,  #PCI_ADDR_EXT]
	@ and return
		mov	r0, #0				@ and return to traps-armv
		mov	pc, r9

		.globl	timer_IRQ_interrupt
timer_IRQ_interrupt:
		mov	r9, lr				@ save lr in case we get interrupted?
							@ this should only ever be IRQ_TIMER0
		mov	r5, #0x10			@ IRQ_TIMER0_MASK
		mov	r4, #IO_BASE_ARM_CSR	
		str	r5, [r4, #IRQ_ENABLE_CLEAR]	@ Disable interrupt source
		ldr	r6, LC0
		ldr	r7, [r6]
		add	r2, r7, #1
		str	r2, [r6]
		mrs	r8, cpsr
		bl	do_IRQ				@ off to C code, which should eventually hit the timer code
		msr	cpsr, r8			@ Restore interrupts
		str	r7, [r6]
		str	r5, [r4, #IRQ_ENABLE_SET]	@ Re-enable interrupt source
		mov	r0, #0
		mov	pc, r9

		.globl	probe_IRQ_interrupt
probe_IRQ_interrupt:
		mov	r9, lr
		@ we should block further irq's of this type
	mov r3, r0
	add r0, r0, #'0'
	bl putchar
	mov r0, #'$'
	bl putchar
	mov r0, r3
		mov	r6, #IO_BASE_ARM_CSR		@ handle footbridge int
		cmp	r0, #16				@ disable this particular interrupt
		bge	1f				@ skip to PIC type interrupts
		ldr	r7, LC1				@ init in irq.c
		ldr	r7, [r7, r0, lsl #2]		@ get the bit mask
		str	r7, [r6, #IRQ_ENABLE_CLEAR]	@ block further ints of this type
		b	2f

1:		mov	r2, #IO_BASE
		cmp	r0, #24
		movge	r5, #PIC_MASK_HI
		subge	r0, r0, #24
		movlt	r5, #PIC_MASK_LO
		sublt	r0, r0, #16
		mov	r4, #1
		mov	r4, r4, lsl r0
		mov	r3, #0x40000000			@ IO work around
//		str	r3, [r6, #PCI_ADDR_EXT]
		ldrb	r8, [r2, r5]
		orr	r4, r8, r4 
		strb	r4, [r2, r5]			@ set the PIC disable mask
		mov	r0, #0
//		str	r0, [r6, #PCI_ADDR_EXT]
		mov	r0, #0x20			@ standard PIC EOI
		strb	r0, [r2, #PIC_HI]
		b	9f
9:		strb	r0, [r2, #PIC_LO]		@ send to both PIC chips
2:
		mov	r0, #0
		mov	pc, r9

		.global	bad_IRQ
bad_IRQ:	mov	r1, r0
		adr	r0, Lmsg
		b	printk

Lmsg:		.ascii	"Bad interrupt %d received!\n\0"
		.align

